@@ -1,7 +1,7 @@
 /* Various processing of names.
 
    Copyright (C) 1988, 1992, 1994, 1996, 1997, 1998, 1999, 2000, 2001,
-   2003, 2004 Free Software Foundation, Inc.
+   2003, 2004, 2005, 2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -15,7 +15,7 @@
 
    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
-   59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 #include <system.h>
 
@@ -66,7 +66,7 @@ uid_to_uname (uid_t uid, char **uname)
 
   if (uid != 0 && uid == cached_no_such_uid)
     {
-      *uname = strdup ("");
+      *uname = xstrdup ("");
       return;
     }
 
@@ -81,11 +81,11 @@ uid_to_uname (uid_t uid, char **uname)
       else
 	{
 	  cached_no_such_uid = uid;
-	  *uname = strdup ("");
+	  *uname = xstrdup ("");
 	  return;
 	}
     }
-  *uname = strdup (cached_uname);
+  *uname = xstrdup (cached_uname);
 }
 
 /* Given GID, find the corresponding GNAME.  */
@@ -96,7 +96,7 @@ gid_to_gname (gid_t gid, char **gname)
 
   if (gid != 0 && gid == cached_no_such_gid)
     {
-      *gname = strdup ("");
+      *gname = xstrdup ("");
       return;
     }
 
@@ -111,11 +111,11 @@ gid_to_gname (gid_t gid, char **gname)
       else
 	{
 	  cached_no_such_gid = gid;
-	  *gname = strdup ("");
+	  *gname = xstrdup ("");
 	  return;
 	}
     }
-  *gname = strdup (cached_gname);
+  *gname = xstrdup (cached_gname);
 }
 
 /* Given UNAME, set the corresponding UID and return 1, or else, return 0.  */
@@ -209,10 +209,10 @@ name_add (const char *name)
     }
   name_array[names++] = name;
 }
+
 
 /* Names from external name file.  */
 
-static FILE *name_file;		/* file to read names from */
 static char *name_buffer;	/* buffer to hold the current file name */
 static size_t name_buffer_length; /* allocated length of name_buffer */
 
@@ -233,17 +233,6 @@ name_init (void)
 {
   name_buffer = xmalloc (NAME_FIELD_SIZE + 2);
   name_buffer_length = NAME_FIELD_SIZE;
-
-  if (files_from_option)
-    {
-      if (!strcmp (files_from_option, "-"))
-	{
-	  request_stdin ("-T");
-	  name_file = stdin;
-	}
-      else if (name_file = fopen (files_from_option, "r"), !name_file)
-	open_fatal (files_from_option);
-    }
 }
 
 void
@@ -253,47 +242,6 @@ name_term (void)
   free (name_array);
 }
 
-/* Read the next filename from name_file and null-terminate it.  Put
-   it into name_buffer, reallocating and adjusting name_buffer_length
-   if necessary.  Return 0 at end of file, 1 otherwise.  */
-static int
-read_name_from_file (void)
-{
-  int character;
-  size_t counter = 0;
-
-  /* FIXME: getc may be called even if character was EOF the last time here.  */
-
-  /* FIXME: This + 2 allocation might serve no purpose.  */
-
-  while (character = getc (name_file),
-	 character != EOF && character != filename_terminator)
-    {
-      if (counter == name_buffer_length)
-	{
-	  if (name_buffer_length * 2 < name_buffer_length)
-	    xalloc_die ();
-	  name_buffer_length *= 2;
-	  name_buffer = xrealloc (name_buffer, name_buffer_length + 2);
-	}
-      name_buffer[counter++] = character;
-    }
-
-  if (counter == 0 && character == EOF)
-    return 0;
-
-  if (counter == name_buffer_length)
-    {
-      if (name_buffer_length * 2 < name_buffer_length)
-	xalloc_die ();
-      name_buffer_length *= 2;
-      name_buffer = xrealloc (name_buffer, name_buffer_length + 2);
-    }
-  name_buffer[counter] = '\0';
-
-  return 1;
-}
-
 /* Get the next name from ARGV or the file of names.  Result is in
    static storage and can't be relied upon across two calls.
 
@@ -311,37 +259,25 @@ name_next (int change_dirs)
   if (filename_terminator == '\0')
     change_dirs = 0;
 
-  while (1)
+  while (name_index != names)
     {
-      /* Get a name, either from file or from saved arguments.  */
-
-      if (name_index == names)
+      size_t source_len;
+      source = name_array[name_index++];
+      source_len = strlen (source);
+      if (name_buffer_length < source_len)
 	{
-	  if (! name_file)
-	    break;
-	  if (! read_name_from_file ())
-	    break;
-	}
-      else
-	{
-	  size_t source_len;
-	  source = name_array[name_index++];
-	  source_len = strlen (source);
-	  if (name_buffer_length < source_len)
+	  do
 	    {
-	      do
-		{
-		  name_buffer_length *= 2;
-		  if (! name_buffer_length)
-		    xalloc_die ();
-		}
-	      while (name_buffer_length < source_len);
-
-	      free (name_buffer);
-	      name_buffer = xmalloc (name_buffer_length + 2);
+	      name_buffer_length *= 2;
+	      if (! name_buffer_length)
+		xalloc_die ();
 	    }
-	  strcpy (name_buffer, source);
+	  while (name_buffer_length < source_len);
+
+	  free (name_buffer);
+	  name_buffer = xmalloc (name_buffer_length + 2);
 	}
+      strcpy (name_buffer, source);
 
       /* Zap trailing slashes.  */
 
@@ -359,30 +295,17 @@ name_next (int change_dirs)
 	chdir_flag = 1;
       else
 	{
-	  unquote_string (name_buffer);
+	  if (unquote_option)
+	    unquote_string (name_buffer);
 	  if (incremental_option)
 	    register_individual_file (name_buffer);
 	  return name_buffer;
 	}
     }
 
-  /* No more names in file.  */
-
-  if (name_file && chdir_flag)
-    FATAL_ERROR ((0, 0, _("Missing file name after -C")));
-
   return 0;
 }
 
-/* Close the name file, if any.  */
-void
-name_close (void)
-{
-  if (name_file && name_file != stdin)
-    if (fclose (name_file) != 0)
-      close_error (name_buffer);
-}
-
 /* Gather names in a list for scanning.  Could hash them later if we
    really care.
 
@@ -506,6 +429,7 @@ addname (char const *string, int change_dir)
   name->firstch = 1;		/* assume first char is literal */
   name->change_dir = change_dir;
   name->dir_contents = 0;
+  name->explicit = 1;
 
   if (string && is_pattern (string))
     {
@@ -522,7 +446,7 @@ addname (char const *string, int change_dir)
 /* Find a match for FILE_NAME (whose string length is LENGTH) in the name
    list.  */
 static struct name *
-namelist_match (char const *file_name, size_t length)
+namelist_match (char const *file_name, size_t length, bool exact)
 {
   struct name *p;
 
@@ -535,6 +459,8 @@ namelist_match (char const *file_name, size_t length)
 
       if (p->regexp
 	  ? fnmatch (p->name, file_name, recursion_option) == 0
+	  : exact ? (p->length == length
+		     && memcmp (file_name, p->name, length) == 0)
 	  : (p->length <= length
 	     && (file_name[p->length] == '\0'
 		 || (ISSLASH (file_name[p->length]) && recursion_option))
@@ -557,17 +483,17 @@ name_match (const char *file_name)
       struct name *cursor = namelist;
 
       if (!cursor)
-	return ! files_from_option;
+	return 1;
 
       if (cursor->fake)
 	{
 	  chdir_do (cursor->change_dir);
 	  namelist = 0;
 	  nametail = &namelist;
-	  return ! files_from_option;
+	  return 1;
 	}
 
-      cursor = namelist_match (file_name, length);
+      cursor = namelist_match (file_name, length, false);
       if (cursor)
 	{
 	  if (!(ISSLASH (file_name[cursor->length]) && recursion_option)
@@ -616,6 +542,8 @@ all_names_found (struct tar_stat_info *p)
   struct name const *cursor;
   size_t len;
 
+  if (test_label_option)
+    return true;
   if (!p->file_name || occurrence_option == 0 || p->had_trailing_slash)
     return false;
   len = strlen (p->file_name);
@@ -784,6 +712,8 @@ add_hierarchy_to_namelist (struct name *name, dev_t device)
 	  string_length = strlen (string);
 	  if (*string == 'D')
 	    {
+	      struct name *np;
+
 	      if (allocated_length <= name_length + string_length)
 		{
 		  do
@@ -797,8 +727,9 @@ add_hierarchy_to_namelist (struct name *name, dev_t device)
 		  namebuf = xrealloc (namebuf, allocated_length + 1);
 		}
 	      strcpy (namebuf + name_length, string + 1);
-	      add_hierarchy_to_namelist (addname (namebuf, change_dir),
-					 device);
+	      np = addname (namebuf, change_dir);
+	      np->explicit = 0;
+	      add_hierarchy_to_namelist (np, device);
 	    }
 	}
 
@@ -858,18 +789,22 @@ collect_and_sort_names (void)
     name->found_count = 0;
 }
 
-/* This is like name_match, except that it returns a pointer to the
-   name it matched, and doesn't set FOUND in structure.  The caller
-   will have to do that if it wants to.  Oh, and if the namelist is
-   empty, it returns null, unlike name_match, which returns TRUE.  */
+/* This is like name_match, except that
+    1. It returns a pointer to the name it matched, and doesn't set FOUND
+    in structure. The caller will have to do that if it wants to.
+    2. If the namelist is empty, it returns null, unlike name_match, which
+    returns TRUE.
+    3. The second argument (EXACT) controls matching algorithm. If it
+    is TRUE, the exact matching is used. However, regular expressions are
+    always matched as such, no matter what the value of EXACT is. */
 struct name *
-name_scan (const char *file_name)
+name_scan (const char *file_name, bool exact)
 {
   size_t length = strlen (file_name);
 
   while (1)
     {
-      struct name *cursor = namelist_match (file_name, length);
+      struct name *cursor = namelist_match (file_name, length, exact);
       if (cursor)
 	return cursor;
 
@@ -939,54 +874,7 @@ new_name (const char *file_name, const char *name)
 bool
 excluded_name (char const *name)
 {
-  return excluded_filename (excluded, name + FILE_SYSTEM_PREFIX_LEN (name));
-}
-
-/* Hash tables of strings.  */
-
-/* Calculate the hash of a string.  */
-static unsigned
-hash_string_hasher (void const *name, unsigned n_buckets)
-{
-  return hash_string (name, n_buckets);
-}
-
-/* Compare two strings for equality.  */
-static bool
-hash_string_compare (void const *name1, void const *name2)
-{
-  return strcmp (name1, name2) == 0;
-}
-
-/* Return zero if TABLE contains a copy of STRING; otherwise, insert a
-   copy of STRING to TABLE and return 1.  */
-static bool
-hash_string_insert (Hash_table **table, char const *string)
-{
-  Hash_table *t = *table;
-  char *s = xstrdup (string);
-  char *e;
-
-  if (! ((t
-	  || (*table = t = hash_initialize (0, 0, hash_string_hasher,
-					    hash_string_compare, 0)))
-	 && (e = hash_insert (t, s))))
-    xalloc_die ();
-
-  if (e == s)
-    return 1;
-  else
-    {
-      free (s);
-      return 0;
-    }
-}
-
-/* Return 1 if TABLE contains STRING.  */
-static bool
-hash_string_lookup (Hash_table const *table, char const *string)
-{
-  return table && hash_lookup (table, string);
+  return excluded_file_name (excluded, name + FILE_SYSTEM_PREFIX_LEN (name));
 }
 
 /* Names to avoid dumping.  */
@@ -1030,92 +918,6 @@ is_individual_file (char const *name)
 
 
 
-static Hash_table *prefix_table[2];
-
-/* Return true if file names of some members in the archive were stripped off
-   their leading components. We could have used
-        return prefix_table[0] || prefix_table[1]
-   but the following seems to be safer: */
-bool
-removed_prefixes_p (void)
-{
-  return (prefix_table[0] && hash_get_n_entries (prefix_table[0]) != 0)
-         || (prefix_table[1] && hash_get_n_entries (prefix_table[1]) != 0);
-}
-
-/* Return a safer suffix of FILE_NAME, or "." if it has no safer
-   suffix.  Check for fully specified file names and other atrocities.
-   Warn the user if we do not return NAME.  If LINK_TARGET is 1,
-   FILE_NAME is the target of a hard link, not a member name.  */
-
-char *
-safer_name_suffix (char const *file_name, bool link_target)
-{
-  char const *p;
-
-  if (absolute_names_option)
-    p = file_name;
-  else
-    {
-      /* Skip file system prefixes, leading file name components that contain
-	 "..", and leading slashes.  */
-
-      size_t prefix_len = FILE_SYSTEM_PREFIX_LEN (file_name);
-
-      for (p = file_name + prefix_len; *p; )
-	{
-          if (p[0] == '.' && p[1] == '.' && (ISSLASH (p[2]) || !p[2]))
-	    prefix_len = p + 2 - file_name;
-
-	  do
-	    {
-	      char c = *p++;
-	      if (ISSLASH (c))
-		break;
-	    }
-	  while (*p);
-	}
-
-      for (p = file_name + prefix_len; ISSLASH (*p); p++)
-	continue;
-      prefix_len = p - file_name;
-
-      if (prefix_len)
-	{
-	  char *prefix = alloca (prefix_len + 1);
-	  memcpy (prefix, file_name, prefix_len);
-	  prefix[prefix_len] = '\0';
-
-	  if (hash_string_insert (&prefix_table[link_target], prefix))
-	    {
-	      static char const *const diagnostic[] =
-	      {
-		N_("Removing leading `%s' from member names"),
-		N_("Removing leading `%s' from hard link targets")
-	      };
-	      WARN ((0, 0, _(diagnostic[link_target]), prefix));
-	    }
-	}
-    }
-
-  if (! *p)
-    {
-      if (p == file_name)
-	{
-	  static char const *const diagnostic[] =
-	  {
-	    N_("Substituting `.' for empty member name"),
-	    N_("Substituting `.' for empty hard link target")
-	  };
-	  WARN ((0, 0, "%s", _(diagnostic[link_target])));
-	}
-
-      p = ".";
-    }
-
-  return (char *) p;
-}
-
 /* Return the size of the prefix of FILE_NAME that is removed after
    stripping NUM leading file name components.  NUM must be
    positive.  */
